import binascii
from automower_ble.helpers import crc
from enum import IntEnum
import asyncio
import logging
import json
from importlib.resources import files
from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak_retry_connector import establish_connection, BleakClientWithServiceCache
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from bleak import BleakClient

logger = logging.getLogger(__name__)


class ModeOfOperation(IntEnum):
    # ProtocolTypes$IMowerAppMowerMode, used in modeOfOperation: 4586, 1
    # Comments from: https://developer.husqvarnagroup.cloud/apis/Automower+Connect+API?tab=status%20description%20and%20error%20codes#user-content-mode
    AUTO = 0
    MANUAL = 1
    HOME = 2  # Mower goes home and parks forever. Week schedule is not used. Cannot be overridden with forced mowing.
    DEMO = 3  # Same as main area, but shorter times. No blade operation
    POI = 4


class MowerState(IntEnum):
    # ProtocolTypes$IMowerAppState, used in mowerState: 4586, 2
    # Comments from: https://developer.husqvarnagroup.cloud/apis/Automower+Connect+API?tab=status%20description%20and%20error%20codes#user-content-state
    OFF = 0  # Mower is turned off.
    WAIT_FOR_SAFETYPIN = 1
    STOPPED = 2  # Mower is stopped requires manual action.
    FATAL_ERROR = 3
    PENDING_START = 4
    PAUSED = 5  # Mower has been paused by user.
    IN_OPERATION = 6  # See value in activity for status.
    RESTRICTED = (
        7  # Mower can currently not mow due to week calender, or override park.
    )
    ERROR = 8  # An error has occurred. Check errorCode. Mower requires manual action.


class MowerActivity(IntEnum):
    # ProtocolTypes$IMowerAppActivity, used in mowerActivity: 4586, 3
    # Comments from: https://developer.husqvarnagroup.cloud/apis/Automower+Connect+API?tab=status%20description%20and%20error%20codes#user-content-activity
    NONE = 0
    CHARGING = 1  # Mower is charging in station due to low battery.
    GOING_OUT = 2
    MOWING = 3  # Mower is mowing lawn. If in demo mode the blades are not in operation.
    GOING_HOME = 4  # Mower is going home to the charging station.
    PARKED = 5
    STOPPED_IN_GARDEN = 6  # Mower has stopped. Needs manual action to resume


class OverrideAction(IntEnum):
    NONE = 0
    FORCEDPARK = 1
    FORCEDMOW = 2


class ResponseResult(IntEnum):
    OK = 0
    UNKNOWN_ERROR = 1
    INVALID_VALUE = 2
    OUT_OF_RANGE = 3
    NOT_AVAILABLE = 4
    NOT_ALLOWED = 5
    INVALID_GROUP = 6
    INVALID_ID = 7
    DEVICE_BUSY = 8
    INVALID_PIN = 9
    MOWER_BLOCKED = 10


class TaskInformation:
    def __init__(
        self,
        next_start_time,
        duration_in_seconds,
        on_monday,
        on_tuesday,
        on_wednesday,
        on_thursday,
        on_friday,
        on_saturday,
        on_sunday,
    ):
        self.next_start_time = next_start_time
        self.duration_in_seconds = duration_in_seconds
        self.on_monday = on_monday
        self.on_tuesday = on_tuesday
        self.on_wednesday = on_wednesday
        self.on_thursday = on_thursday
        self.on_friday = on_friday
        self.on_saturday = on_saturday
        self.on_sunday = on_sunday


class Command:
    def __init__(self, channel_id: int, parameter: dict):
        self.channel_id = channel_id

        self.major = parameter["major"]
        self.minor = parameter["minor"]

        self.request_data_type = parameter.get("requestType")

        if "responseType" not in parameter:
            parameter["responseType"] = "no_response"

        if not isinstance(parameter["responseType"], dict):  # Always wrap in list
            self.response_data_type = {"response": parameter["responseType"]}
        else:
            self.response_data_type = parameter["responseType"]
        self.request_data = bytearray()

    def generate_request(self, **kwargs) -> bytearray:
        self.request_data = bytearray(18)
        self.request_data[0] = 0x02  # Hard coded value (start of packet)
        self.request_data[1] = 0xFD  # 0xFD = LINKED_PACKET_TYPE
        self.request_data[2] = 0x00  # Length, low byte, updated later
        self.request_data[3] = 0x00  # Length, high byte, updated later

        # ChannelID
        self.request_data[4:8] = self.channel_id.to_bytes(4, byteorder="little")

        self.request_data[8] = 0x01  # is_linked (usually 0x01)

        self.request_data[9] = 0x00  # CRC, Updated later
        self.request_data[10] = (
            0x00  # Packet type (0x00 = request, 0x01 = response, 0x02 = event)
        )
        self.request_data[11] = 0xAF  # Hard coded value

        major_bytes = self.major.to_bytes(2, byteorder="little")

        self.request_data[12] = major_bytes[0]  # low byte of 'module'
        self.request_data[13] = major_bytes[1]  # high byte of 'module'
        self.request_data[14] = self.minor  # low byte of 'command'
        self.request_data[15] = 0x00  # high byte of 'command'

        # Byte 16 represents length of request data type
        request_length = 0
        request_data = bytearray()
        if self.request_data_type is not None:
            for request_name, request_type in self.request_data_type.items():
                if request_name not in kwargs:
                    raise ValueError(
                        "Missing request parameter: "
                        + request_name
                        + " for command ("
                        + str(self.major)
                        + ", "
                        + str(self.minor)
                        + ")"
                    )

                if request_type == "uint32":
                    request_length += 4
                    request_data += kwargs[request_name].to_bytes(4, byteorder="little")
                elif request_type == "uint16":
                    request_length += 2
                    request_data += kwargs[request_name].to_bytes(2, byteorder="little")
                elif request_type == "uint8":
                    request_length += 1
                    request_data += kwargs[request_name].to_bytes(1, byteorder="little")
                else:
                    raise ValueError("Unknown request type: " + request_type)
        self.request_data[16] = request_length

        self.request_data[17] = 0x00  # high byte of request_length
        if request_length > 0:
            self.request_data += request_data

        self.request_data[2] = len(self.request_data) - 2  # Length

        self.request_data[9] = crc(self.request_data, 1, 8)  # CRC

        # Two last bytes are crc and 0x03
        self.request_data.append(crc(self.request_data, 1, len(self.request_data) - 1))
        self.request_data.append(0x03)  # Hard coded value

        return self.request_data

    def parse_response(self, response_data: bytearray) -> dict[str, int | str] | None:
        response_length = response_data[17]
        data = response_data[19 : 19 + response_length]
        response: dict[str, int | str] = {}
        dpos = 0  # data position
        for name, dtype in self.response_data_type.items():
            if dtype == "no_response":
                return None
            if (dtype == "tUnixTime") or (dtype == "uint32"):
                response[name] = int.from_bytes(
                    data[dpos : dpos + 4], byteorder="little"
                )
                dpos += 4
            elif dtype == "uint16":
                response[name] = int.from_bytes(
                    data[dpos : dpos + 2], byteorder="little"
                )
                dpos += 2
            elif (dtype == "uint8") or (dtype == "bool"):
                response[name] = data[dpos]
                dpos += 1
            elif dtype == "ascii":
                if len(self.response_data_type) != 1:
                    raise ValueError(
                        "ASCII response type can currently only be used when there is only one response type"
                    )
                response[name] = data.decode("ascii").rstrip(
                    "\x00"
                )  # Remove trailing null bytes
                dpos += len(data)
            else:
                raise ValueError("Unknown data type: " + dtype)
        if dpos != len(data):
            raise ValueError(f"Data length mismatch. Read {dpos} bytes of {len(data)}")
        return response

    def validate_command_response(self, response_data: bytearray) -> bool:
        if response_data[0] != 0x02:
            return False

        if response_data[1] != 0xFD:
            return False

        if response_data[3] != 0x00:  # high byte of length
            return False

        if response_data[4:8] != self.channel_id.to_bytes(4, byteorder="little"):
            return False

        if response_data[8] != 0x01:
            # This is a valid config, but we don't support it
            # return m1656b(decodeState, c10786f);
            return False

        if response_data[9] != crc(response_data, 1, 8):
            return False

        if response_data[10] != 0x01:  # packet type is not 0x01 = response
            return False

        if response_data[11] != 0xAF:
            return False

        major_bytes = self.major.to_bytes(4, byteorder="little")
        if response_data[12] != major_bytes[0]:
            return False
        if response_data[13] != major_bytes[1]:
            return False
        if response_data[14] != self.minor:
            return False

        if response_data[15] != 0x00:  # high byte of 'command' (self.minor)
            return False

        if (
            response_data[16] != 0x00
        ):  # result: OK(0), UNKNOWN_ERROR(1), INVALID_VALUE(2), OUT_OF_RANGE(3), NOT_AVAILABLE(4), NOT_ALLOWED(5), INVALID_GROUP(6), INVALID_ID(7), DEVICE_BUSY(8), INVALID_PIN(9), MOWER_BLOCKED(10);
            logger.warning("Non zero response result: %d", response_data[16])
            return False

        return True


class BLEClient:
    def __init__(self, channel_id: int, address, pin=None):
        self.channel_id = channel_id
        self.address = address
        self.pin = pin
        self.MTU_SIZE = 20

        self.lock = asyncio.Lock()
        self.queue: asyncio.Queue[bytearray] = asyncio.Queue()

        self.client: BleakClient | None = None
        self.protocol = None

    async def get_protocol(self):
        if self.protocol is None:

            def read_protocol_file():
                with files("automower_ble").joinpath("protocol.json").open("r") as f:
                    return json.load(f)

            self.protocol = await asyncio.get_running_loop().run_in_executor(
                None, read_protocol_file
            )
        return self.protocol

    async def _get_response(self):
        try:
            data = await asyncio.wait_for(self.queue.get(), timeout=10)

        except TimeoutError:
            logger.error("Unable to get response from device: '%s'", self.address)
            return None

        return data

    async def _write_data(self, data):
        logger.info("Writing: %s", str(binascii.hexlify(data)))

        chunk_size = self.MTU_SIZE - 3
        for chunk in (
            data[i : i + chunk_size] for i in range(0, len(data), chunk_size)
        ):
            await self.client.write_gatt_char(self.write_char, chunk, response=False)

        logger.debug("Finished writing")

    async def _read_data(self):
        data = await self._get_response()

        if data is None:
            return None

        if len(data) < 3:
            # We got such a small amount of data, let's try again
            if (chunk := await self._get_response()) is None:
                return None
            data = data + chunk

            if len(data) < 3:
                # Something is wrong
                return None

        length = data[2] + 4

        logger.debug("Waiting for %d bytes", length)

        while len(data) < length:
            try:
                data = data + await asyncio.wait_for(self.queue.get(), timeout=5)
            except TimeoutError:
                logger.error(
                    "Unable to get full response from device: '%s', currently have %s",
                    str(binascii.hexlify(data)),
                    self.address,
                )
                logger.error("Expecting %d bytes, only have %d", length, len(data))
                return None

        logger.info("Final response: %s", str(binascii.hexlify(data)))

        return data

    async def _request_response(self, request_data):
        async with self.lock:
            try:
                # If there are previous responses, flush them out
                while not self.queue.empty():
                    await self.queue.get()

                await self._write_data(request_data)

                response_data = await self._read_data()
                if response_data is None:
                    logger.error(
                        "Unable to communicate with device: '%s'", self.address
                    )
                    if self.is_connected():
                        await self.disconnect()
                    return None

            except asyncio.exceptions.CancelledError:
                logger.debug("Received CancelledError")
                if self.is_connected():
                    await self.disconnect()
                return None

        return response_data

    async def connect(self, device) -> ResponseResult:
        """
        Connect to a device and setup the channel

        Returns a ResponseResult
        """
        logger.info("starting scan...")

        if device is None:
            logger.error("could not find device with address '%s'", self.address)
            return ResponseResult.UNKNOWN_ERROR

        logger.info("connecting to device...")
        self.client = await establish_connection(
            BleakClientWithServiceCache,
            device,
            device.name or "Unknown Device",
        )
        logger.info("connected")

        logger.info("pairing device...")
        try:
            await self.client.pair()
            logger.info("paired")
        except Exception as err:
            logger.debug("Pairing not completed, continuing anyway: %s", err)

        # This is not safe, _mtu_size is not defined in BaseBleakClient but may
        # be defined in subclasses.
        self.client._backend._mtu_size = self.MTU_SIZE  # type: ignore[attr-defined]

        for service in self.client.services:
            logger.info("[Service] %s", service)

            for char in service.characteristics:
                if (
                    "read" in char.properties
                    and char.uuid != "98bd0003-0b0e-421a-84e5-ddbf75dc6de4"
                ):
                    try:
                        value = await self.client.read_gatt_char(char.uuid)
                        logger.debug(
                            "  [Characteristic] %s (%s), Value: %r",
                            char,
                            ",".join(char.properties),
                            value,
                        )
                    except Exception as e:
                        logger.error(
                            "  [Characteristic] %s (%s), Error: %s",
                            char,
                            ",".join(char.properties),
                            e,
                        )
                else:
                    logger.debug(
                        "  [Characteristic] %s (%s)", char, ",".join(char.properties)
                    )
                if char.uuid == "98bd0002-0b0e-421a-84e5-ddbf75dc6de4":
                    self.write_char = char

                if char.uuid == "98bd0003-0b0e-421a-84e5-ddbf75dc6de4":
                    self.read_char = char

        async def notification_handler(
            characteristic: BleakGATTCharacteristic, data: bytearray
        ):
            logger.info("Received: %s", str(binascii.hexlify(data)))
            await self.queue.put(data)

        try:
            await self.client.start_notify(self.read_char, notification_handler)
        except Exception:
            await self.client.disconnect()
            raise

        await asyncio.sleep(5.0)

        request = self.generate_request_setup_channel_id()
        response = await self._request_response(request)
        if response is None:
            return ResponseResult.UNKNOWN_ERROR

        request = self.generate_request_handshake()
        response = await self._request_response(request)
        if response is None:
            return ResponseResult.UNKNOWN_ERROR

        if self.pin is not None:
            command = Command(
                self.channel_id, (await self.get_protocol())["EnterOperatorPin"]
            )
            request = command.generate_request(code=self.pin)
            response = await self._request_response(request)
            if response is None:
                return ResponseResult.UNKNOWN_ERROR
            result = self.get_response_result(response)
            # If the result is UNKNOWN_ERROR, assume the pin was invalid
            return (
                ResponseResult.INVALID_PIN
                if result == ResponseResult.UNKNOWN_ERROR
                else result
            )

        return ResponseResult.OK

    def is_connected(self) -> bool:
        return bool(self.client and self.client.is_connected)

    async def probe_gatts(self, device):
        logger.info("connecting to device...")
        client = await establish_connection(
            BleakClientWithServiceCache,
            device,
            device.name or "Unknown Device",
            max_attempts=3,  # Will retry up to 3 times with backoff
        )
        logger.info("connected")

        manufacture = None
        model = None
        device_type = None

        for service in client.services:
            logger.debug("[Service] %s", service)

            if service.uuid == "98bd0001-0b0e-421a-84e5-ddbf75dc6de4":
                manufacture = service.description

            for char in service.characteristics:
                if "read" in char.properties:
                    try:
                        value = await client.read_gatt_char(char.uuid)
                        logger.debug(
                            "  [Characteristic] %s (%s), Value: %r",
                            char,
                            ",".join(char.properties),
                            value,
                        )
                        if char.uuid == "00002a00-0000-1000-8000-00805f9b34fb":
                            model = value.decode()
                        if char.uuid == "98bd0004-0b0e-421a-84e5-ddbf75dc6de4":
                            device_type = value.rstrip(b"\x00").decode()
                    except Exception as e:
                        logger.error(
                            "  [Characteristic] %s (%s), Error: %s",
                            char,
                            ",".join(char.properties),
                            e,
                        )
                else:
                    logger.debug(
                        "  [Characteristic] %s (%s)", char, ",".join(char.properties)
                    )

        await client.disconnect()

        return (manufacture, device_type, model)

    async def disconnect(self):
        """
        Disconnect from the mower, this should be called after every
        `connect()` before the Python script exits
        """

        await self.client.stop_notify(self.read_char)
        await self.queue.put(None)

        logger.info("disconnecting...")
        await self.client.disconnect()
        logger.info("disconnected")

    def generate_request_setup_channel_id(self) -> bytearray:
        """
        Setup the channelID with an Automower, this is the first
        command that should be sent
        """
        data = bytearray.fromhex("02fd160000000000002e1400000000000000004d61696e00")

        # New ChannelID
        data[11:15] = self.channel_id.to_bytes(4, byteorder="little")

        # CRC and end byte
        data[9] = crc(data, 1, 8)
        data.append(crc(data, 1, len(data) - 1))
        data.append(0x03)

        return data

    def generate_request_handshake(self) -> bytearray:
        """
        Generate a request handshake. This should be called after
        the channel id is set up but before other commands
        """
        data = bytearray.fromhex("02fd0a000000000000d00801")

        data[4:8] = self.channel_id.to_bytes(4, byteorder="little")

        # CRCs and end byte
        data[9] = crc(data, 1, 8)
        data.append(crc(data, 1, len(data) - 1))
        data.append(0x03)

        return data

    def validate_response(self, response_data: bytearray) -> bool:
        if response_data[0] != 0x02:
            return False

        if response_data[1] != 0xFD:
            return False

        if response_data[3] != 0x00:  # high byte of length
            return False

        if response_data[4:8] != self.channel_id.to_bytes(4, byteorder="little"):
            return False

        if response_data[8] != 0x01:
            # This is a valid config, but we don't support it
            # return m1656b(decodeState, c10786f);
            return False

        if response_data[9] != crc(response_data, 1, 8):
            return False

        if response_data[10] != 0x01:  # packet type is not 0x01 = response
            return False

        if response_data[11] != 0xAF:
            return False

        if (
            response_data[16] != 0x00
        ):  # result: OK(0), UNKNOWN_ERROR(1), INVALID_VALUE(2), OUT_OF_RANGE(3), NOT_AVAILABLE(4), NOT_ALLOWED(5), INVALID_GROUP(6), INVALID_ID(7), DEVICE_BUSY(8), INVALID_PIN(9), MOWER_BLOCKED(10);
            logger.warning("Non zero response result: %d", response_data[16])
            return False

        return True

    def get_response_result(self, response_data: bytearray) -> ResponseResult:
        if self.validate_response(response_data) is False:
            # Just log if the response is invalid as this has been seen with user
            # logs from official apps. I.e. it is somewhat expected.
            logger.warning("Response failed validation")

        return ResponseResult(response_data[16])



import unittest
import json
from importlib.resources import files
from automower_ble.protocol import BLEClient, Command, ModeOfOperation
import binascii


class TestRequestMethods(unittest.TestCase):
    def setUp(self):
        with files("automower_ble").joinpath("protocol.json").open("r") as f:
            self.protocol = json.load(f)  # Load the parameters to have them available

    def test_generate_request_setup_channel_id(self):
        client_one = BLEClient(1197489075, "00:00:00:00:00:00")
        client_two = BLEClient(1739453030, "00:00:00:00:00:00")

        self.assertEqual(
            binascii.hexlify(client_one.generate_request_setup_channel_id()),
            b"02fd160000000000002e14b33b6047000000004d61696e001b03",
        )
        self.assertEqual(
            binascii.hexlify(client_one.generate_request_handshake()),
            b"02fd0a00b33b6047005d08012803",
        )

        self.assertEqual(
            binascii.hexlify(client_two.generate_request_setup_channel_id()),
            b"02fd160000000000002e1466f2ad67000000004d61696e003403",
        )
        self.assertEqual(
            binascii.hexlify(client_two.generate_request_handshake()),
            b"02fd0a0066f2ad6700f908012803",
        )

    def test_generate_request_device_type(self):
        command = Command(1739453030, parameter=self.protocol["GetModel"])

        self.assertEqual(
            binascii.hexlify(command.generate_request()),
            b"02fd100066f2ad6701d700af5a1209000000b703",
        )

    def test_generate_request_pin(self):
        command = Command(1739453030, parameter=self.protocol["EnterOperatorPin"])

        self.assertEqual(
            binascii.hexlify(command.generate_request(code=7201)),
            b"02fd120066f2ad6701ad00af381204000200211c7d03",
        )

    def test_generate_request_remaining_charge_time(self):
        command = Command(
            1197489075, parameter=self.protocol["GetRemainingChargingTime"]
        )

        self.assertEqual(
            binascii.hexlify(command.generate_request()),
            b"02fd1000b33b6047017300af0a1016000000b803",
        )

    def test_generate_request_is_charging(self):
        command = Command(1197489075, parameter=self.protocol["IsCharging"])

        self.assertEqual(
            binascii.hexlify(command.generate_request()),
            b"02fd1000b33b6047017300af0a10150000003003",
        )

    def test_generate_request_battery_level(self):
        command = Command(1197489075, parameter=self.protocol["GetBatteryLevel"])

        self.assertEqual(
            binascii.hexlify(command.generate_request()),
            b"02fd1000b33b6047017300af0a1014000000bf03",
        )

    def test_generate_request_set_mode_of_operation(self):
        command = Command(0x5798CA1A, parameter=self.protocol["SetMode"])

        self.assertEqual(
            binascii.hexlify(command.generate_request(mode=ModeOfOperation.AUTO.value)),
            b"02fd11001aca9857013400afea1100000100003303",
        )

    def test_generate_request_get_mode_of_operation(self):
        command = Command(0x5798CA1A, parameter=self.protocol["GetMode"])

        self.assertEqual(
            binascii.hexlify(command.generate_request()),
            b"02fd10001aca9857010900afea1101000000e203",
        )

    def test_generate_request_override_duration(self):
        command = Command(0x5798CA1A, parameter=self.protocol["SetOverrideMow"])

        self.assertEqual(
            binascii.hexlify(command.generate_request(duration=3 * 3600)),
            b"02fd14001aca985701fd00af321203000400302a00004603",
        )

    def test_generate_get_task_request(self):
        command = Command(0x13A51453, parameter=self.protocol["GetTask"])

        self.assertEqual(
            binascii.hexlify(command.generate_request(taskId=0)),
            b"02fd14005314a513019d00af52120500040000000000d203",
        )

    def test_generate_get_number_of_tasks_request(self):
        command = Command(0x13A51453, parameter=self.protocol["GetNumberOfTasks"])

        self.assertEqual(
            binascii.hexlify(command.generate_request()),
            b"02fd10005314a513016900af5212040000008b03",
        )

    def test_generate_get_restriction_reason(self):
        command = Command(0x13A51453, parameter=self.protocol["GetRestrictionReason"])

        self.assertEqual(
            binascii.hexlify(command.generate_request()),
            b"02fd10005314a513016900af3212000000004603",
        )

    def test_generate_get_serial_number(self):
        command = Command(0x13A51453, parameter=self.protocol["GetSerialNumber"])

        self.assertEqual(
            binascii.hexlify(command.generate_request()),
            b"02fd10005314a513016900af5a120a0000003f03",
        )

    def test_generate_is_operator_loggedin_request(self):
        command = Command(0x13A51453, parameter=self.protocol["IsOperatorLoggedIn"])

        self.assertEqual(
            binascii.hexlify(command.generate_request()),
            b"02fd10005314a513016900af3812030000000103",
        )

    def test_generate_getStartupSequenceRequiredRequest(self):
        command = Command(
            0x13A51453, parameter=self.protocol["GetStartupSequenceRequired"]
        )

        self.assertEqual(
            binascii.hexlify(command.generate_request()),
            b"02fd10005314a513016900af5a12020000002303",
        )

    def test_generate_keepalive_request(self):
        command = Command(0x13A51453, parameter=self.protocol["KeepAlive"])

        self.assertEqual(
            binascii.hexlify(command.generate_request()),
            b"02fd10005314a513016900af421202000000d903",
        )

    def test_generate_request_trigger_request(self):
        command = Command(0x13A51453, parameter=self.protocol["StartTrigger"])

        self.assertEqual(
            binascii.hexlify(command.generate_request()),
            b"02fd10005314a513016900afea11040000006303",
        )

    def test_generate_request_override(self):
        command = Command(0x13A51453, parameter=self.protocol["GetOverride"])

        self.assertEqual(
            binascii.hexlify(command.generate_request()),
            b"02fd10005314a513016900af3212020000004103",
        )


if __name__ == "__main__":
    unittest.main()



"""
The top level script to connect and communicate with the mower
This sends requests and decodes responses. This is an example of
how the request and response classes can be used.
"""

# Copyright: Alistair Francis <alistair@alistair23.me>

import argparse
import asyncio
import datetime as dt
import logging

from automower_ble.protocol import (
    BLEClient,
    Command,
    MowerState,
    MowerActivity,
    ModeOfOperation,
    ResponseResult,
    TaskInformation,
)
from automower_ble.models import MowerModels
from automower_ble.error_codes import ErrorCodes

from bleak import BleakScanner

logger = logging.getLogger(__name__)


class Mower(BLEClient):
    def __init__(self, channel_id: int, address, pin=None):
        super().__init__(channel_id, address, pin)
        self.keep_alive_event = asyncio.Event()

    async def connect(self, device) -> ResponseResult:
        """
        Connect to a device and setup the channel

        Returns a ResponseResult
        """
        status = await super().connect(device)
        if status == ResponseResult.OK:
            self.task = asyncio.create_task(self._keep_alive())
        return status

    async def disconnect(self):
        """
        Disconnect from the mower, this should be called after every
        `connect()` before the Python script exits
        """
        self.keep_alive_event.set()
        return await super().disconnect()

    async def _keep_alive(self):
        """
        Keep the connection alive by sending a request every 15 seconds.
        This is needed to prevent the connection from being closed by the mower.
        """
        while not self.keep_alive_event.is_set():
            try:
                if self.is_connected():
                    logger.debug("Sending keep alive")
                    await self.command("KeepAlive")
            except Exception as e:
                logger.warning("Failed to send keep alive: %s", e)
            await asyncio.sleep(15)

    async def command(self, command_name: str, **kwargs):
        """
        This function is used to simplify the communication of the mower using the commands found in protocol.json.
        It will send a request to the mower and then wait for a response. The response will be parsed and returned to the caller.
        """
        command = Command(self.channel_id, (await self.get_protocol())[command_name])
        request = command.generate_request(**kwargs)
        response = await self._request_response(request)
        if response is None:
            return None

        if command.validate_command_response(response) is False:
            # Just log if the response is invalid as this has been seen with user
            # logs from official apps. I.e. it is somewhat expected.
            logger.warning("Response failed validation")

        response_dict = command.parse_response(response)
        if (
            response_dict is not None and len(response_dict) == 1
        ):  # If there is only one key in the response, return the value
            return response_dict["response"]
        return response_dict

    async def get_manufacturer(self) -> str | None:
        """Get the mower manufacturer"""
        model = await self.command("GetModel")
        if model is None:
            return None

        model_information = MowerModels.get(
            (model["deviceType"], model["deviceVariant"])
        )
        if model_information is None:
            return f"Unknown Manufacturer ({model['deviceType']}, {model['deviceVariant']})"

        return model_information.manufacturer

    async def get_model(self) -> str | None:
        """Get the mower model"""
        model = await self.command("GetModel")
        if model is None:
            return None

        model_information = MowerModels.get(
            (model["deviceType"], model["deviceVariant"])
        )
        if model_information is None:
            return f"Unknown Model ({model['deviceType']}, {model['deviceVariant']})"

        return model_information.model

    async def is_charging(self) -> bool:
        """Get the mower charging status"""
        return bool(await self.command("IsCharging"))

    async def battery_level(self) -> int | None:
        """Query the mower battery level"""
        return await self.command("GetBatteryLevel")

    async def mower_state(self) -> MowerState | None:
        """Query the mower state"""
        state = await self.command("GetState")
        if state is None:
            return None
        return MowerState(state)

    async def mower_next_start_time(self) -> dt.datetime | None:
        """Query the mower next start time"""
        next_start_time = await self.command("GetNextStartTime")
        if next_start_time is None or next_start_time == 0:
            return None
        return dt.datetime.fromtimestamp(next_start_time, dt.UTC)

    async def mower_activity(self) -> MowerActivity | None:
        """Query the mower activity"""
        activity = await self.command("GetActivity")
        if activity is None:
            return None
        return MowerActivity(activity)

    async def mower_override(self, duration_hours: float = 3.0) -> None:
        """
        Force the mower to run for the specified duration in hours.
        """
        if duration_hours <= 0:
            raise ValueError("Duration must be greater than 0")

        # Set mode of operation to auto:
        await self.command("SetMode", mode=ModeOfOperation.AUTO)

        # Set the duration of operation:
        await self.command("SetOverrideMow", duration=int(duration_hours * 3600))

        # Request trigger to start, the response validation is expected to fail
        await self.command("StartTrigger")

    async def mower_pause(self):
        await self.command("Pause")

    async def mower_resume(self):
        # The response validation is expected to fail
        await self.command("StartTrigger")

    async def mower_park(self):
        await self.command("SetOverrideParkUntilNextStart")

        # Request trigger to start, the response validation is expected to fail
        await self.command("StartTrigger")

    async def get_task(self, taskid: int) -> TaskInformation | None:
        """
        Get information about a specific task
        """
        task = await self.command("GetTask", taskId=taskid)
        if task is None:
            return None
        return TaskInformation(
            task["start"],
            task["duration"],
            task["useOnMonday"],
            task["useOnTuesday"],
            task["useOnWednesday"],
            task["useOnThursday"],
            task["useOnFriday"],
            task["useOnSaturday"],
            task["useOnSunday"],
        )


async def main(mower: Mower):
    device = await BleakScanner.find_device_by_address(mower.address)

    if device is None:
        print("Unable to connect to device address: " + mower.address)
        print(
            "Please make sure the device address is correct, the device is powered on and nearby"
        )
        return

    await mower.connect(device)

    manufacturer = await mower.get_manufacturer()
    print("Mower manufacturer: " + (manufacturer or "Unknown manufacturer"))

    model = await mower.get_model()
    print("Mower model: " + (model or "Unknown model"))

    charging = await mower.is_charging()
    if charging:
        print("Mower is charging")
    else:
        print("Mower is not charging")

    battery_level = await mower.battery_level()
    print("Battery is: " + str(battery_level) + "%")

    state = await mower.mower_state()
    if state is not None:
        print("Mower state: " + state.name)

    activity = await mower.mower_activity()
    if activity is not None:
        print("Mower activity: " + activity.name)

    next_start_time = await mower.mower_next_start_time()
    if next_start_time:
        print("Next start time: " + next_start_time.strftime("%Y-%m-%d %H:%M:%S"))
    else:
        print("No next start time")

    statuses = await mower.command("GetAllStatistics")
    for status, value in statuses.items():
        print(status, value)

    serial_number = await mower.command("GetSerialNumber")
    print("Serial number: " + str(serial_number))

    mower_name = await mower.command("GetUserMowerNameAsAsciiString")
    print("Mower name: " + mower_name)

    # print("Running for 3 hours")
    # await mower.mower_override()

    # print("Pause")
    # await mower.mower_pause()

    # print("Resume")
    # await mower.mower_resume()

    # activity = await mower.mower_activity()
    # print("Mower activity: " + activity)

    # If command argument passed then send command
    if args.command:
        print("Sending command to control mower (" + args.command + ")")
        match args.command:
            case "park":
                print("command=park")
                cmd_result = await mower.mower_park()
            case "pause":
                print("command=pause")
                cmd_result = await mower.mower_pause()
            case "resume":
                print("command=resume")
                cmd_result = await mower.mower_resume()
            case "override":
                print("command=override")
                cmd_result = await mower.mower_override()  # type: ignore[func-returns-value]
            case _:
                print("command=??? (Unknown command: " + args.command + ")")
        print("command result = " + str(cmd_result))

    # moved last message after command, this seems to cause all future commands/queries to fail
    last_message = await mower.command("GetMessage", messageId=0)
    print("Last message: ")
    print(
        "\t"
        + dt.datetime.fromtimestamp(last_message["time"], dt.UTC).strftime(
            "%Y-%m-%d %H:%M:%S"
        )
    )
    print("\t" + ErrorCodes(last_message["code"]).name)

    await mower.disconnect()


if __name__ == "__main__":
    parser = argparse.ArgumentParser()

    device_group = parser.add_mutually_exclusive_group(required=True)

    device_group.add_argument(
        "--address",
        metavar="<address>",
        help="the Bluetooth address of the Automower device to connect to",
    )

    parser.add_argument(
        "--pin",
        metavar="<code>",
        type=int,
        default=None,
        help="Send PIN to authenticate. This feature is experimental and might not work.",
    )

    parser.add_argument(
        "--command",
        metavar="<command>",
        default=None,
        help="Send command to control mower (one of resume, pause, park or override)",
    )

    args = parser.parse_args()

    mower = Mower(1197489078, args.address, args.pin)

    log_level = logging.INFO
    logging.basicConfig(
        level=log_level,
        format="%(asctime)-15s %(name)-8s %(levelname)s: %(message)s",
    )

    asyncio.run(main(mower))



import unittest
import json
from importlib.resources import files
from automower_ble.protocol import Command, MowerState, MowerActivity
from automower_ble.models import MowerModels


class TestRequestMethods(unittest.TestCase):
    def setUp(self):
        with files("automower_ble").joinpath("protocol.json").open("r") as f:
            self.protocol = json.load(f)  # Load the parameters to have them available

    def test_decode_response_device_type(self):
        command = Command(1197489078, parameter=self.protocol["GetModel"])

        response = command.parse_response(
            bytearray.fromhex("02fd1300b63b604701e601af5a1209000002001701c803")
        )
        self.assertEqual(
            MowerModels[(response["deviceType"], response["deviceVariant"])].model,
            "Automower 305",
        )

        response = command.parse_response(
            bytearray.fromhex("02fd130038e38f0b01dc01af5a1209000002000c005903")
        )
        self.assertEqual(
            MowerModels[(response["deviceType"], response["deviceVariant"])].model,
            "Automower 315",
        )

    def test_decode_response_is_charging(self):
        command = Command(1197489078, self.protocol["IsCharging"])

        self.assertEqual(
            command.parse_response(
                bytearray.fromhex("02fd1200b63b604701db01af0a101500000100011603")
            )["response"],
            True,
        )
        self.assertEqual(
            command.parse_response(
                bytearray.fromhex("02fd1200b63b604701db01af0a101500000100004803")
            )["response"],
            False,
        )

    def test_decode_response_mower_state(self):
        command = Command(1197489078, self.protocol["GetState"])

        self.assertNotIn(
            command.parse_response(
                bytearray.fromhex("02fd1200b33b6047010901afea110100000100008103")
            )["response"],
            [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 12, 13, 14],  # 11=unknown
        )

        command = Command(876143061, self.protocol["GetState"])
        self.assertEqual(
            command.parse_response(
                bytearray.fromhex("02fd1200d5e13834012301afea110200000100033a03")
            )["response"],
            MowerState.FATAL_ERROR.value,
        )

    def test_decode_response_mower_activity(self):
        response = Command(1197489078, self.protocol["GetActivity"])

        self.assertEqual(
            response.parse_response(
                bytearray.fromhex("02fd1200b33b6047010901afea110200000100026403")
            )["response"],
            MowerActivity.GOING_OUT.value,  # 2 = goingOut
        )

    def test_decode_get_task_response(self):
        response = Command(1197489078, self.protocol["GetTask"])
        decoded = response.parse_response(
            bytearray.fromhex(
                "02fd240025be246a010701af5212050000130000e1000038310000010001010001013003"
            )
        )

        self.assertEqual(
            decoded["start"],
            57600,
        )
        self.assertEqual(
            decoded["duration"],
            12600,
        )
        self.assertEqual(decoded["useOnMonday"], 1)
        self.assertEqual(decoded["useOnTuesday"], 0)
        self.assertEqual(decoded["useOnWednesday"], 1)
        self.assertEqual(decoded["useOnThursday"], 1)
        self.assertEqual(decoded["useOnFriday"], 0)
        self.assertEqual(decoded["useOnSaturday"], 1)
        self.assertEqual(decoded["useOnSunday"], 1)

    def test_decode_get_number_of_tasks_response(self):
        response = Command(0x13A51453, self.protocol["GetNumberOfTasks"])
        self.assertEqual(
            response.parse_response(
                bytearray.fromhex("02fd150025be246a012e01af52120400000400010000004f03")
            )["response"],
            1,
        )


if __name__ == "__main__":
    unittest.main()



"""
Adopted from the example provided by the Bleak library:
https://github.com/hbldh/bleak/blob/develop/examples/discover.py
"""

import argparse
import asyncio

from bleak import BleakScanner

# "Husqvarna AB" maps to "0x0426" for the manufacturer data
# Please see: https://bitbucket.org/bluetooth-SIG/public/src/main/assigned_numbers/company_identifiers/company_identifiers.yaml
HUSQVARNA_COMPANY_IDENTIFIER = 0x0426


async def main(args: argparse.Namespace):
    print(f"Scanning for {args.timeout} seconds, please wait...")

    devices = await BleakScanner.discover(
        timeout=args.timeout,
        return_adv=True,
        cb={"use_bdaddr": args.macos_use_bdaddr},
    )

    husqvarna_device_found = False
    for d, a in devices.values():
        if (
            args.show_all
            or next(iter(a.manufacturer_data.keys()), None)
            == HUSQVARNA_COMPANY_IDENTIFIER
        ):
            if not husqvarna_device_found and not args.show_all:
                print("Husqvarna device(s) found!")
                husqvarna_device_found = True

            print(f"\n\tAddress: {d.address}")
            print(f"\tName: {d.name}")
            print(f"\tSignal Strength: {a.rssi} dBm (closer to 0 is stronger)")
            if args.show_all:
                print(f"\tManufacturer Data: {a.manufacturer_data}")

    if not husqvarna_device_found and not args.show_all:
        print("No Husqvarna devices found!")
        print("Make sure your Automower is powered on and nearby.")


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--macos-use-bdaddr",
        action="store_true",
        help="when true use Bluetooth address instead of UUID on macOS",
    )
    parser.add_argument(
        "--timeout",
        type=float,
        default=15.0,
        help="Duration (seconds) to scan for BLE devices. Default = 15 seconds",
    )
    parser.add_argument(
        "--show-all",
        action="store_true",
        help="Scan and show all devices found, not just only Husqvarna devices.",
    )
    args = parser.parse_args()
    asyncio.run(main(args))



"""
A list of supported Automower models
"""


# Copyright: Alistair Francis <alistair@alistair23.me>
class ModelInformation:
    def __init__(self, manufacturer: str, model: str):
        self.manufacturer = manufacturer
        self.model = model


MowerModels = {
    (0, 0): ModelInformation("Unknown", "Unknown (0,0)"),
    (5, 0): ModelInformation("Husqvarna", "Automower 420"),
    (7, 0): ModelInformation("Husqvarna", "Automower 430X"),
    (7, 1): ModelInformation("Husqvarna", "Automower 430XH"),
    (8, 0): ModelInformation("Husqvarna", "Automower 450X"),
    (8, 1): ModelInformation("Husqvarna", "Automower 450XH"),
    (11, 0): ModelInformation("Husqvarna", "Automower 310"),
    (12, 0): ModelInformation("Husqvarna", "Automower 315"),
    (12, 1): ModelInformation("Husqvarna", "Automower 315X"),
    (12, 23): ModelInformation("Husqvarna", "Automower 315X Limited Edition"),
    (14, 1): ModelInformation("Gardena", "SILENO City 250"),
    (14, 2): ModelInformation("Gardena", "SILENO City 300"),
    (14, 3): ModelInformation("Gardena", "SILENO City Unknown (14,3)"),
    (14, 4): ModelInformation("Gardena", "SILENO City Unknown (14,4)"),
    (14, 5): ModelInformation("Gardena", "SILENO City Unknown (14,5)"),
    (14, 6): ModelInformation("Gardena", "SILENO City 500"),
    (14, 7): ModelInformation("Gardena", "Smart SILENO City Unknown (14,7)"),
    (14, 8): ModelInformation("Gardena", "Smart SILENO City Unknown (14,8)"),
    (14, 9): ModelInformation("Gardena", "SILENO City Unknown (14,9)"),
    (14, 10): ModelInformation("Gardena", "SILENO City Unknown (14,10)"),
    (14, 11): ModelInformation("Gardena", "SILENO City Unknown (14,11)"),
    (14, 12): ModelInformation("Gardena", "Smart SILENO City Unknown (14,12)"),
    (14, 13): ModelInformation("Gardena", "SILENO City Unknown (14,13)"),
    (14, 14): ModelInformation("Gardena", "Smart SILENO City Unknown (14,14)"),
    (14, 15): ModelInformation("Gardena", "Smart SILENO City Unknown (14,15)"),
    (14, 16): ModelInformation("Gardena", "Smart SILENO City Unknown (14,16)"),
    (14, 17): ModelInformation("Gardena", "Smart SILENO City Unknown (14,17)"),
    (14, 18): ModelInformation("Gardena", "SILENO City Unknown (14,18)"),
    (15, 0): ModelInformation("Husqvarna", "Automower 440"),
    (16, 0): ModelInformation("Husqvarna", "Automower 550"),
    (16, 1): ModelInformation("Husqvarna", "Automower 550H"),
    (17, 0): ModelInformation("Husqvarna", "Automower 520"),
    (17, 1): ModelInformation("Husqvarna", "Automower 520H"),
    (18, 1): ModelInformation("Gardena", "SILENO Life 750"),
    (18, 2): ModelInformation("Gardena", "SILENO Life Unknown (18,2)"),
    (18, 3): ModelInformation("Gardena", "SILENO Life Unknown (18,3)"),
    (18, 4): ModelInformation("Gardena", "Smart SILENO Life Unknown (18,4)"),
    (18, 5): ModelInformation("Gardena", "Smart SILENO Life Unknown (18,5)"),
    (18, 6): ModelInformation("Gardena", "Smart SILENO Life Unknown (18,6)"),
    (18, 7): ModelInformation("Gardena", "SILENO Life Unknown (18,7)"),
    (18, 8): ModelInformation("Gardena", "SILENO Life Unknown (18,8)"),
    (18, 9): ModelInformation("Gardena", "Smart SILENO Life Unknown (18,9)"),
    (18, 10): ModelInformation("Gardena", "Smart SILENO Life Unknown (18,10)"),
    (18, 11): ModelInformation("Gardena", "Smart SILENO Life Unknown (18,11)"),
    (18, 12): ModelInformation("Gardena", "SILENO Life Unknown (18,12)"),
    (18, 13): ModelInformation("Gardena", "SILENO Life Unknown (18,13)"),
    (18, 14): ModelInformation("Gardena", "Smart SILENO Life Unknown (18,14)"),
    (18, 15): ModelInformation("Gardena", "Smart SILENO Life Unknown (18,15)"),
    (18, 16): ModelInformation("Gardena", "SILENO Life Unknown (18,16)"),
    (18, 17): ModelInformation("Gardena", "Smart SILENO Life Unknown (18,17)"),
    (19, 1): ModelInformation("Husqvarna", "Automower 115H"),
    (20, 0): ModelInformation("Husqvarna", "Automower 435X AWD"),
    (21, 0): ModelInformation("Husqvarna", "Automower 535 AWD"),
    (22, 1): ModelInformation("McCulloch", "Rob S400"),
    (22, 2): ModelInformation("McCulloch", "Rob S600"),
    (22, 3): ModelInformation("McCulloch", "Rob Unknown (22,3)"),
    (22, 4): ModelInformation("McCulloch", "Rob S800"),
    (23, 1): ModelInformation("Husqvarna", "Automower 305"),
    (23, 2): ModelInformation("Husqvarna", "Automower 310 Mark II"),
    (23, 3): ModelInformation("Husqvarna", "Automower 315 Mark II"),
    (24, 1): ModelInformation("Husqvarna", "Automower Ceora 546 EPOS"),
    (24, 2): ModelInformation("Husqvarna", "Automower Ceora 544 EPOS"),
    (25, 1): ModelInformation("Flymo", "Easilife Unknown (25,1)"),
    (25, 2): ModelInformation("Flymo", "Easilife Unknown (25,2)"),
    (25, 3): ModelInformation("Flymo", "Easilife Unknown (25,3)"),
    (25, 4): ModelInformation("Flymo", "Easilife Unknown (25,4)"),
    (26, 1): ModelInformation("Husqvarna", "Automower 550 EPOS"),
    (26, 2): ModelInformation("Husqvarna", "Automower 550H EPOS"),
    (26, 3): ModelInformation("Husqvarna", "Automower 520 EPOS"),
    (26, 4): ModelInformation("Husqvarna", "Automower 520H EPOS"),
    (27, 1): ModelInformation("Husqvarna", "Liberty Pilot EPAC"),
    (27, 2): ModelInformation("Husqvarna", "Liberty Pilot NA"),
    (27, 3): ModelInformation("Husqvarna", "Liberty Pilot HiCut"),
    (28, 1): ModelInformation("Husqvarna", "Automower 405X"),
    (28, 2): ModelInformation("Husqvarna", "Automower 415X"),
    (29, 1): ModelInformation("Gardena", "SILENO Minimo Unknown (29,1)"),
    (29, 2): ModelInformation("Gardena", "SILENO Minimo 400"),
    (29, 3): ModelInformation("Gardena", "SILENO Minimo 500"),
    (29, 4): ModelInformation("Gardena", "SILENO Minimo 250"),
    (29, 5): ModelInformation("Gardena", "SILENO Minimo Unknown (29,5)"),
    (29, 6): ModelInformation("Gardena", "SILENO Minimo Unknown (29,6)"),
    (29, 7): ModelInformation("Gardena", "SILENO Minimo Unknown (29,7)"),
    (29, 8): ModelInformation("Gardena", "SILENO Minimo Unknown (29,8)"),
    (29, 9): ModelInformation("Gardena", "SILENO Minimo 350"),
    (29, 10): ModelInformation("Gardena", "SILENO Minimo Unknown (29,10)"),
    (30, 1): ModelInformation("Flymo", "Easilife Go 200"),
    (30, 2): ModelInformation("Flymo", "Easilife Go Unknown (30,2)"),
    (30, 3): ModelInformation("Flymo", "Easilife Go 150"),
    (30, 4): ModelInformation("Flymo", "Easilife Go 250 (30,4)"),
    (30, 5): ModelInformation("Flymo", "Easilife Go 500"),
    (31, 1): ModelInformation("Husqvarna", "Automower 320 Nera"),
    (31, 2): ModelInformation("Husqvarna", "Automower 430X Nera"),
    (31, 3): ModelInformation("Husqvarna", "Automower 450X Nera"),
    (32, 1): ModelInformation("Husqvarna", "Automower Aspire R4"),
    (33, 1): ModelInformation("Husqvarna", "Automower 430X Egalite"),
    (39, 2): ModelInformation("Husqvarna", "Automower 310E Nera"),
    (39, 5): ModelInformation("Husqvarna", "Automower 410XE Nera"),
    (40, 1): ModelInformation("Husqvarna", "Automower 450X EPOS"),
    (40, 2): ModelInformation("Husqvarna", "Automower 450XH EPOS"),
}
